"""
Implementation of the XDG Menu Specification Version 1.0
http://www.freedesktop.org/standards/menu-spec/

Not Supported (and not planed):
    - <LegacyDir>, <KDELegacyDirs>

TODO:
	- Maybe a bit performance tweeking
"""

from xdg.BaseDirectory import *
from xdg.DesktopEntry import *
from Exceptions import *

import re, os, xml.dom.minidom

class Menu:
	"Final Menu Class"
	def __init__(self):
		# Public attributes
		self.Name = ""
		self.Directory = []
		self.DeskEntries = []
		self.Submenus = []

		# Private attributes, only needed for parsing
		self.AppDirs = []
		self.DirectoryDirs = []
		self.OnlyUnallocated = False
		self.Deleted = False
		self.Rules = []
		self.Move = []

	def __str__(self):
		return self.Name

	def __add__(self, other):
		for dir in other.AppDirs:
			if not dir in self.AppDirs:
				self.AppDirs.append(dir)
		for dir in other.DirectoryDirs:
			if not dir in self.DirectoryDirs:
				self.DirectoryDirs.append(dir)
		for entry in other.Directory:
			if not entry in self.Directory:
				self.Directory.append(entry)
		for osubmenu in other.Submenus:
			ssubmenu = self.get_submenu(osubmenu.Name)
			if ssubmenu:
				ssubmenu = ssubmenu + osubmenu
				break
			else:
				self.Submenus.append(osubmenu)

		for rule in other.Rules:
			self.Rules.append(rule)

		self.OnlyUnallocated = other.OnlyUnallocated
		self.Deleted = other.Deleted

		return self
	
	def replace_move(self, newmove):
		for move in self.Move:
			if newmove[0] == move[0]:
				move == newmove
				return True
		else:
			return False
	
	def get_submenu(self, newmenu):
		for submenu in self.Submenus:
			if submenu.Name == newmenu:
				return submenu
		else:
			return False

class Rule:
	"Inlcude / Exclude Rules Class"
	def __init__(self):
		# Public attributes
		self.Type = ''
		self.Rule = ''

		# Private attributes, only needed for parsing
		self.Depth = -1
		self.Expr = [ ]
		self.New = True


class MenuParser:
	Root = Menu()
	depth = -1
	path = "/"
	basename = ""
	__mergeFiles = []

	def parse(self, file = "", wm = "None"):
		# check if file exists
		if not file:
			dirs = BaseDirectory().get_XDG_CONFIG_DIRS()
			print dirs
			for dir in dirs:
				file = dir + "/menus/applications.menu"
				if os.path.isdir(dir):
					if os.path.isfile(file):
						break
			else:
				raise ParsingError("File found", file)

		# convert to absolute path
		if not os.path.isabs(file):
			file = os.path.abspath(file)

		# check if it is a .menu file
		if not re.search('\.menu$', file):
			raise ParsingError('Not a .menu file', file)

		# create xml parser
		try:
			doc = xml.dom.minidom.parse(file)
		except IOError:
			raise ParsingError('File not found', file)
		except xml.parsers.expat.ExpatError:
			raise ParsingError('Not a valid .menu file', file)

		# save path of the menu entrie in self.path
		(self.path, name) = os.path.split(file)
		(self.basename, None) = name.split(".")

		# parse menufile
		self.__preparse(doc)
		self.__parse(doc, self.Root)
		self.__postparse(self.Root)

		# finally generate the menu
		cache = DesktopEntryCache(wm)
		self.__genmenuNotOnlyAllocated(self.Root, cache)
		self.__genmenuOnlyAllocated(self.Root, cache)

		return self.Root


	def __preparse(self, doc):
		# replace default dir stuff
		for name in [ "DefaultAppDirs", "DefaultDirectoryDirs", "DefaultMergeDirs" ]:
			if name == "DefaultMergeDirs":
				dirs = BaseDirectory().get_XDG_CONFIG_DIRS()
			else:
				dirs = BaseDirectory().get_XDG_DATA_DIRS()
			dirs.reverse()

			nodes = doc.getElementsByTagName(name)
			for node in nodes:
				parent = node.parentNode
				for dir in dirs:
					entry = doc.createElement(re.sub("Default", "", name).replace("Dirs", "Dir"))
					if name == "DefaultAppDirs":
						entry.appendChild(doc.createTextNode(dir + "/applications"))
					elif name == "DefaultDirectoryDirs":
						entry.appendChild(doc.createTextNode(dir + "/desktop-directories"))
					elif name == "DefaultMergeDirs":
						entry.appendChild(doc.createTextNode(dir + "/" + self.basename + "-parsed"))
					parent.insertBefore(entry, node)
				parent.removeChild(node)

		# Remove duplicate stuff, convert to absolute path and remove not existing path
		for name in [ "MergeDir", "MergeFile", "DirectoryDir", "AppDir" ]:
			values  = []
			parents = []
			nodes = doc.getElementsByTagName(name)
			for node in nodes:
				if not os.path.isabs(node.childNodes[0].nodeValue):
					node.childNodes[0].nodeValue = os.path.join(self.path, node.childNodes[0].nodeValue)
				if ( node.childNodes[0].nodeValue in values and node.parentNode in parents ) \
				or ( not os.path.exists(node.childNodes[0].nodeValue) ) \
				or ( name != "MergeFile" and not os.path.isdir(node.childNodes[0].nodeValue) ) \
				or ( name == "MergeFile" and not os.path.isfile(node.childNodes[0].nodeValue) ):
					node.parentNode.removeChild(node)
				else:
					parents.append(node.parentNode)
					values.append(node.childNodes[0].nodeValue)


	def __parse(self, node, parent):
		for child in node.childNodes:
			if child.nodeType == xml.dom.Node.ELEMENT_NODE:
				if child.tagName == 'Menu':
					self.__parseMenu(child, parent)
				elif child.tagName == 'AppDir':
					self.__parseAppDir(child.childNodes[0].nodeValue, parent)
				elif child.tagName == 'DirectoryDir':
					self.__parseDirectoryDir(child.childNodes[0].nodeValue, parent)
				elif child.tagName == 'Name' :
					self.__parseName(child.childNodes[0].nodeValue, parent)
				elif child.tagName == 'Directory' :
					self.__parseDirectory(child.childNodes[0].nodeValue, parent)
				elif child.tagName == 'OnlyUnallocated':
					self.__parseOnlyUnallocated(None, parent)
				elif child.tagName == 'NotOnlyUnallocated':
					self.__parseNotOnlyUnallocated(None, parent)
				elif child.tagName == 'Deleted':
					self.__parseDeleted(None, parent)
				elif child.tagName == 'NotDeleted':
					self.__parseNotDeleted(None, parent)
				elif child.tagName == 'Include':
					self.__parseInclude(child, parent)
				elif child.tagName == 'Exclude':
					self.__parseExclude(child, parent)
				elif child.tagName == 'Filename':
					self.__parseFilename(child.childNodes[0].nodeValue, parent)
				elif child.tagName == 'Category':
					self.__parseCategory(child.childNodes[0].nodeValue, parent)
				elif child.tagName == 'All':
					self.__parseAll(None, parent)
				elif child.tagName == 'And':
					self.__parseAnd(child, parent)
				elif child.tagName == 'Or':
					self.__parseOr(child, parent)
				elif child.tagName == 'Not':
					self.__parseNot(child, parent)
				elif child.tagName == 'MergeFile':
					self.__parseMergeFile(child)
				elif child.tagName == 'MergeDir':
					self.__parseMergeDir(child)
				elif child.tagName == 'MergeEnd':
					self.__parseMergeEnd()
				elif child.tagName == 'Move':
					self.__parseMove(child, parent)

	def __postparse(self, menu):
		# parse move operations
		for move in menu.Move:
			if menu.get_submenu(move[0]):
				move_from_menu = menu.get_submenu(move[0])
				move_to_menu = menu.get_submenu(move[1])
				if move_to_menu:
					move_to_menu = move_to_menu + move_from_menu
					menu.Submenus.remove(move_from_menu)
				else:
					move_from_menu.Name = move[1]

		# go recursive through all menus
		for submenu in menu.Submenus:
			# delete <Delete> Menus
			if submenu.Deleted == True:
				menu.Submenus.remove(submenu)
				continue
			# add parent's app/directory dirs
			for dir in menu.AppDirs:
				if not dir in submenu.AppDirs:
					submenu.AppDirs.insert(0, dir)
			for dir in menu.DirectoryDirs:
				if not dir in submenu.DirectoryDirs:
					submenu.DirectoryDirs.insert(0, dir)

			# get the valid .directory file out of the list
			entry = ""
			for directory in submenu.Directory:
				for dir in submenu.DirectoryDirs:
					if os.path.exists(os.path.join(dir, directory)):
						try:
							entry = DesktopEntry()
							entry.parse(os.path.join(dir, directory))
						except:
							pass
			if entry:
				submenu.Directory = entry
			else:
				submenu.Directory = DesktopEntry()

			# enter submenus
			self.__postparse(submenu)

	def __parseMenu(self, node, parent):
		self.depth += 1
		m = Menu()
		self.__parse(node, m)
		if self.depth == 0:
			self.Root = m
		else:
			submenu = parent.get_submenu(m.Name)
			if submenu:
				submenu = submenu + m
			else:
				parent.Submenus.append(m)
		self.depth -= 1

	def __parseAppDir(self, dir, parent):
		parent.AppDirs.append(dir)

	def __parseDirectoryDir(self, dir, parent):
		parent.DirectoryDirs.append(dir)

	def __parseName(self, value, parent):
		parent.Name = value

	def __parseDirectory(self, value, parent):
		parent.Directory.append(value)

	def __parseOnlyUnallocated(self, None, parent):
		parent.OnlyUnallocated = True

	def __parseNotOnlyUnallocated(self, None, parent):
		parent.OnlyUnallocated = False

	def __parseDeleted(self, None, parent):
		parent.Deleted = True

	def __parseNotDeleted(self, None, parent):
		parent.Deleted = False

	# include / exclude rule parsing
	def __parseInclude(self, node, parent):
		rule = Rule()
		rule.Type = 'Include'
		rule.Depth += 1
		rule.Expr.append(" or ")
		parent.Rules.append(rule)
		self.__parse(node, parent)
		rule.Depth -= 1

	def __parseExclude(self, node, parent):
		rule = Rule()
		rule.Type = 'Exclude'
		rule.Depth += 1
		rule.Expr.append(" or ")
		parent.Rules.append(rule)
		self.__parse(nore, parent)
		rule.Depth -= 1

	def __parseFilename(self, value, parent):
		if not parent.Rules[-1].New:
			parent.Rules[-1].Rule += parent.Rules[-1].Expr[parent.Rules[-1].Depth]
		else:
			parent.Rules[-1].New = False
		parent.Rules[-1].Rule += 'path == \'' + value + '\''

	def __parseCategory(self, value, parent):
		if not parent.Rules[-1].New:
			parent.Rules[-1].Rule += parent.Rules[-1].Expr[parent.Rules[-1].Depth]
		else:
			parent.Rules[-1].New = False
		parent.Rules[-1].Rule += '\'' + value + '\' in entry.getCategories()'

	def __parseAll(self, None, parent):
		rule = ''
		if not parent.Rules[-1].New:
			rule += parent.Rules[-1].Expr[parent.Rules[-1].Depth]
		else:
			parent.Rules[-1].New = False
		rule += 'True'
		parent.Rules[-1].Rule += rule

	def __parseAnd(self, node, parent):
		if not parent.Rules[-1].New:
			parent.Rules[-1].Rule += parent.Rules[-1].Expr[parent.Rules[-1].Depth]
		parent.Rules[-1].New = True
		parent.Rules[-1].Rule += '('
		parent.Rules[-1].Depth += 1
		parent.Rules[-1].Expr.append(" and ")
		self.__parse(node, parent)
		parent.Rules[-1].Depth -= 1
		parent.Rules[-1].Expr.pop()
		parent.Rules[-1].Rule += ')'
	
	def __parseOr(self, node, parent):
		if not parent.Rules[-1].New:
			parent.Rules[-1].Rule += parent.Rules[-1].Expr[parent.Rules[-1].Depth]
		parent.Rules[-1].New = True
		parent.Rules[-1].Rule += '('
		parent.Rules[-1].Depth += 1
		parent.Rules[-1].Expr.append(" or ")
		self.__parse(node, parent)
		parent.Rules[-1].Depth -= 1
		parent.Rules[-1].Expr.pop()
		parent.Rules[-1].Rule += ')'

	def __parseNot(self, node, parent):
		if not parent.Rules[-1].New:
			parent.Rules[-1].Rule += parent.Rules[-1].Expr[parent.Rules[-1].Depth]
		parent.Rules[-1].New = True
		parent.Rules[-1].Rule += 'not ('
		parent.Rules[-1].Depth += 1
		parent.Rules[-1].Expr.append(" or ")
		self.__parse(node, parent)
		parent.Rules[-1].Depth -= 1
		parent.Rules[-1].Expr.pop()
		parent.Rules[-1].Rule += ')'

	# Merge Stuff
	def __parseMergeFile(self, child):
		self.__mergeFile(child.childNodes[0].nodeValue, child)

	def __parseMergeDir(self, child):
		files = os.listdir(child.childNodes[0].nodeValue)
		for file in files:
			if re.search('\.menu', file):
				self.__mergeFile(file, child)

	def __mergeFile(self, file, child):
		# check for infinite loops
		if file in self.__mergeFiles:
			raise ParsingError('Infinite MergeFile loop detected', file)

		self.__mergeFiles.append(file)

		# append doc
		try:
			doc = xml.dom.minidom.parse(file)
			self.__preparse(doc)
			next = child.nextSibling
			for node in doc.childNodes[1].childNodes:
				if node.nodeType == xml.dom.Node.ELEMENT_NODE:
					if node.tagName != "Name":
						child.parentNode.insertBefore(node, next)

		except xml.parsers.expat.ExpatError:
			raise ParsingError('Not a valid .menu file', file)

		# Make endnode to pop __mergeFiles again
		endnode = doc.createElement("MergeEnd")
		child.parentNode.insertBefore(endnode, next)

	def __parseMergeEnd(self):
		self.__mergeFiles.pop()

	def __parseMove(self, child, parent):
		for node in child.childNodes:
			if node.nodeType == xml.dom.Node.ELEMENT_NODE:
				if node.tagName == "Old":
					newmove = []
					newmove.append(node.childNodes[0].nodeValue)
				else:
					newmove.append(node.childNodes[0].nodeValue)
					if not parent.replace_move(newmove):
						parent.Move.append(newmove)

	# Finally generate the menu
	def __genmenuNotOnlyAllocated(self, menu, cache):
		if menu.OnlyUnallocated == False:
			cache.addEntries(menu.AppDirs)
			list = cache.getEntries(menu.AppDirs)
			for path, entry in list.items():
				inc = False
				for rule in menu.Rules:
					if eval(rule.Rule):
						if rule.Type == "Include":
							inc = True
						else:
							inc = False
				if inc == True:
					entry.allocated = True
					menu.DeskEntries.append(entry)

		for submenu in menu.Submenus:
			self.__genmenuNotOnlyAllocated(submenu, cache)

	def __genmenuOnlyAllocated(self, menu, cache):
		if menu.OnlyUnallocated == True:
			cache.addEntries(menu.AppDirs)
			list = cache.getEntries(menu.AppDirs)
			for path, entry in list.items():
				if entry.allocated == True:
					continue
				inc = False
				for rule in menu.Rules:
					if eval(rule.Rule):
						if rule.Type == "Include":
							inc = True
						else:
							inc = False
				if inc == True:
					menu.DeskEntries.append(entry)

		for submenu in menu.Submenus:
			self.__genmenuOnlyAllocated(submenu, cache)
